
#import "Primitive.h"
#import "VScheme.h"
#import "SchemeDelegate.h"

BOOL isBoolean(id item)
{
    return [item isKindOfClass:[Boolean class]];
}

BOOL isChar(id item)
{
    return [item isKindOfClass:[Char class]];
}

BOOL isInt(id item)
{
    return [item isKindOfClass:[Int class]];
}

BOOL isDouble(id item)
{
    return [item isKindOfClass:[Double class]];
}

BOOL isSymbol(id item)
{
    return [item isKindOfClass:[Symbol class]];
}

BOOL isString(id item)
{
    return [item isKindOfClass:[String class]];
}

BOOL isPair(id item)
{
    return [item isKindOfClass:[Pair class]];
}

BOOL isVector(id item)
{
    return [item isKindOfClass:[Vector class]];
}

BOOL isTriple(id item)
{
    return [item isKindOfClass:[Triple class]];
}

BOOL isPrimitive(id item)
{
    return [item isKindOfClass:[Primitive class]];
}

BOOL isClosure(id item)
{
    return [item isKindOfClass:[Closure class]];
}

BOOL isThunk(id item)
{
    return [item isKindOfClass:[Thunk class]];
}

BOOL isFalse(id item)
{
    if(item==[NSNull null]){
        return YES;
    }

    if([item isKindOfClass:[Boolean class]]){
        return ([item boolVal]==YES ? NO : YES);
    }

    return NO;
}

BOOL isEqual(id itema, id itemb)
{
    if([itema class]!=[itemb class]){
        return NO;
    }

    if(isChar(itema)){
        return ([itema charVal]==[itemb charVal] ? YES : NO);
    }
    else if(isInt(itema)){
        return ([itema intVal]==[itemb intVal] ? YES : NO);
    }
    else if(isDouble(itema)){
        return ([itema doubleVal]==[itemb doubleVal] ? YES : NO);
    }
    else if(isBoolean(itema)){
        return ([itema boolVal]==[itemb boolVal] ? YES : NO);
    }
    else if(isString(itema)){
        return [[itema strVal] isEqual:[itemb strVal]];
    }
    else if(isSymbol(itema)){
        return [[itema symVal] isEqual:[itemb symVal]];
    }

    return (itema==itemb ? YES : NO);
}


@implementation Primitive

- init
{
    [super init];

    value = [NSNull null];
    errmsg = nil;

    return self;
}


- (NUMTYPE)checkArgsNumeric:(NSMutableArray *)args offset:(int)offs
{
    int pos;
    NUMTYPE res = NT_INTEGERS;

    for(pos=offs; pos<[args count]; pos++){
        id arg = [args objectAtIndex:pos];
        if(isInt(arg)==NO){
            if(isDouble(arg)==NO){
                return NT_OTHER;
            }
            res = NT_DOUBLE;
        }
    }

    return res;
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
  [self notImplemented: _cmd];
  return NO;
}

- (NSString *)primName
{
  [self notImplemented: _cmd];
  return @"_not_implemented";
}

- value
{
    return value;
}

- errmsg
{
    return errmsg;
}

@end

@implementation PRMVectorPred

- (NSString *)primName
{
    return @"vector?";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id item;

    if(offs+1!=[args count]){
        errmsg = @"vector? takes one argument";
        return NO;
    }

    item = [args objectAtIndex:offs];

    value = [[Boolean alloc] initSCMBoolean:isVector(item)];
    return YES;
}

@end

@implementation PRMPairPred

- (NSString *)primName
{
    return @"pair?";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id item;

    if(offs+1!=[args count]){
        errmsg = @"pair? takes one argument";
        return NO;
    }

    item = [args objectAtIndex:offs];

    value = [[Boolean alloc] initSCMBoolean:isPair(item)];
    return YES;
}

@end

@implementation PRMNullPred

- (NSString *)primName
{
    return @"null?";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id item;
    
    if(offs+1!=[args count]){
        errmsg = @"null? takes one argument";
        return NO;
    }

    item = [args objectAtIndex:offs];

    value = [[Boolean alloc] 
                initSCMBoolean:(item==[NSNull null] ? YES : NO)];
    return YES;
}

@end

@implementation PRMZeroPred

- (NSString *)primName
{
    return @"zero?";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    id item;
    BOOL res = NO;

    
    if(offs+1!=[args count]){
        errmsg = @"zero? takes one argument";
        return NO;
    }

    item = [args objectAtIndex:offs];

    if(nt==NT_OTHER){
        errmsg = @"Argument to zero? must be numeric";
        return NO;
    }

    if(nt==NT_INTEGERS){
        if(![item intVal]){
            res = YES;
        }
    }
    else{
        if([item doubleVal]==(double)0.0){
            res = YES;
        }
    }

    value = [[Boolean alloc] initSCMBoolean:res];
    return YES;
}

@end

@implementation PRMNumberPred

- (NSString *)primName
{
    return @"number?";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    id item;
    BOOL res = NO;

    
    if(offs+1!=[args count]){
        errmsg = @"number? takes one argument";
        return NO;
    }

    item = [args objectAtIndex:offs];

    res = (nt!=NT_OTHER ? YES : NO);

    value = [[Boolean alloc] initSCMBoolean:res];
    return YES;
}

@end

@implementation PRMEqPred

- (NSString *)primName
{
    return @"eq?";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id itema, itemb;
    BOOL res = NO;
    
    if(offs+2!=[args count]){
        errmsg = @"eq? takes two arguments";
        return NO;
    }

    itema = [args objectAtIndex:offs];
    itemb = [args objectAtIndex:offs+1];

    value = [[Boolean alloc] initSCMBoolean:isEqual(itema, itemb)];
    return YES;
}

@end

@implementation PRMNot

- (NSString *)primName
{
    return @"not";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id item;
    
    if(offs+1!=[args count]){
        errmsg = @"not takes one argument";
        return NO;
    }

    item = [args objectAtIndex:offs];

    value = [[Boolean alloc] initSCMBoolean:isFalse(item)];
    return YES;
}

@end

@implementation PRMNumEqual

- (NSString *)primName
{
    return @"=";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    int icur, ires = 0; double dcur, dres = 0;
    int pos = offs;
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    BOOL res = YES;

    if(nt==NT_OTHER){
        errmsg = @"Arguments to = must be numeric";
        return NO;
    }

    if(nt==NT_INTEGERS){
        ires = [[args objectAtIndex:offs] intVal];
        for(pos=offs+1; pos<[args count]; pos++){
            icur = [[args objectAtIndex:pos] intVal];
            if(!(ires==icur)){
                res = NO;
                break;
            }
            ires = icur;
        }
    }
    else{
        dres = [[args objectAtIndex:offs] doubleVal];
        for(pos=offs+1; pos<[args count]; pos++){
            dcur = [[args objectAtIndex:pos] doubleVal];
            if(!(dres==dcur)){
                res = NO;
                break;
            }
            dres = dcur;
        }
    }

    value = [[Boolean alloc] initSCMBoolean:res];
    return YES;
}

@end

@implementation PRMNumLT

- (NSString *)primName
{
    return @"<";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    int icur, ires = 0; double dcur, dres = 0;
    int pos = offs;
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    BOOL res = YES;

    if(nt==NT_OTHER){
        errmsg = @"Arguments to < must be numeric";
        return NO;
    }

    if(nt==NT_INTEGERS){
        ires = [[args objectAtIndex:offs] intVal];
        for(pos=offs+1; pos<[args count]; pos++){
            icur = [[args objectAtIndex:pos] intVal];
            if(!(ires<icur)){
                res = NO;
                break;
            }
            ires = icur;
        }
    }
    else{
        dres = [[args objectAtIndex:offs] doubleVal];
        for(pos=offs+1; pos<[args count]; pos++){
            dcur = [[args objectAtIndex:pos] doubleVal];
            if(!(dres<dcur)){
                res = NO;
                break;
            }
            dres = dcur;
        }
    }

    value = [[Boolean alloc] initSCMBoolean:res];
    return YES;
}

@end

@implementation PRMNumGT

- (NSString *)primName
{
    return @">";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    int icur, ires = 0; double dcur, dres = 0;
    int pos = offs;
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    BOOL res = YES;

    if(nt==NT_OTHER){
        errmsg = @"Arguments to > must be numeric";
        return NO;
    }

    if(nt==NT_INTEGERS){
        ires = [[args objectAtIndex:offs] intVal];
        for(pos=offs+1; pos<[args count]; pos++){
            icur = [[args objectAtIndex:pos] intVal];
            if(!(ires>icur)){
                res = NO;
                break;
            }
            ires = icur;
        }
    }
    else{
        dres = [[args objectAtIndex:offs] doubleVal];
        for(pos=offs+1; pos<[args count]; pos++){
            dcur = [[args objectAtIndex:pos] doubleVal];
            if(!(dres>dcur)){
                res = NO;
                break;
            }
            dres = dcur;
        }
    }

    value = [[Boolean alloc] initSCMBoolean:res];
    return YES;
}

@end

@implementation PRMPlus

- (NSString *)primName
{
    return @"+";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    int ires = 0; double dres = 0;
    int pos = offs;
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];

    if(nt==NT_OTHER){
        errmsg = @"Arguments to + must be numeric";
        return NO;
    }

    if(nt==NT_INTEGERS){
        for(pos=offs; pos<[args count]; pos++){
            ires += [[args objectAtIndex:pos] intVal];
        }

        value = [[Int alloc] initSCMInt:ires];
        return YES;
    }
    
    for(pos=offs; pos<[args count]; pos++){
        dres += [[args objectAtIndex:pos] doubleVal];
    }
    
    value = [[Double alloc] initSCMDouble:dres];
    return YES;
}

@end

@implementation PRMTimes

- (NSString *)primName
{
    return @"*";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    int ires = 1; double dres = 1;
    int pos = offs;
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];

    if(nt==NT_OTHER){
        errmsg = @"Arguments to * must be numeric";
        return NO;
    }

    if(nt==NT_INTEGERS){
        for(pos=offs; pos<[args count]; pos++){
            ires *= [[args objectAtIndex:pos] intVal];
        }

        value = [[Int alloc] initSCMInt:ires];
        return YES;
    }
    
    for(pos=offs; pos<[args count]; pos++){
        dres *= [[args objectAtIndex:pos] doubleVal];
    }
    
    value = [[Double alloc] initSCMDouble:dres];
    return YES;
}

@end

@implementation PRMMinus

- (NSString *)primName
{
    return @"-";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    int ires = 1; double dres = 1;
    int pos = offs;
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];

    if(offs==[args count]){
        errmsg = @"- needs at least one argument";
        return NO;
    }

    if(nt==NT_OTHER){
        errmsg = @"Arguments to - must be numeric";
        return NO;
    }

    if(nt==NT_INTEGERS){
        ires = [[args objectAtIndex:offs] intVal];
        if(offs+1==[args count]){
            ires = -ires;
        }
        else{
            for(pos=offs+1; pos<[args count]; pos++){
                ires -= [[args objectAtIndex:pos] intVal];
            }
        }
        value = [[Int alloc] initSCMInt:ires];
        return YES;
    }
    
    dres = [[args objectAtIndex:offs] doubleVal];
    if(offs+1==[args count]){
        dres = -dres;
    }
    else{
        for(pos=offs+1; pos<[args count]; pos++){
            dres -= [[args objectAtIndex:pos] doubleVal];
        }
    }
    value = [[Double alloc] initSCMDouble:dres];
    return YES;
}

@end

@implementation PRMDivide

- (NSString *)primName
{
    return @"/";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    int ires = 1; double dres = 1;
    int pos = offs;
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];

    if(offs==[args count]){
        errmsg = @"/ needs at least one argument";
        return NO;
    }

    if(nt==NT_OTHER){
        errmsg = @"Arguments to / must be numeric";
        return NO;
    }

    if(nt==NT_INTEGERS){
        ires = [[args objectAtIndex:offs] intVal];
        if(offs+1==[args count]){
            if(ires!=1){
                errmsg = @"no integer fractions";
                return NO;
            }
        }
        else{
            for(pos=offs+1; pos<[args count]; pos++){
                ires /= [[args objectAtIndex:pos] intVal];
            }
        }
        value = [[Int alloc] initSCMInt:ires];
        return YES;
    }
    
    dres = [[args objectAtIndex:offs] doubleVal];
    if(offs+1==[args count]){
        dres = ((double)1.0)/dres;
    }
    else{
        for(pos=offs+1; pos<[args count]; pos++){
            dres /= [[args objectAtIndex:pos] doubleVal];
        }
    }
    value = [[Double alloc] initSCMDouble:dres];
    return YES;
}

@end

@implementation PRMQuotient

- (NSString *)primName
{
    return @"quotient";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    int a, b;

    if(offs+2!=[args count]){
        errmsg = @"quotient takes exactly two arguments";
        return NO;
    }

    if(nt!=NT_INTEGERS){
        errmsg = @"Arguments to quotient must be integers";
        return NO;
    }

    a = [[args objectAtIndex:offs] intVal];
    b = [[args objectAtIndex:offs+1] intVal];

    if(!b){
        errmsg = @"divide by zero error in quotient";
        return NO;
    }

    value = [[Int alloc] initSCMInt:(a/b)];
    return YES;
}

@end

@implementation PRMRemainder

- (NSString *)primName
{
    return @"remainder";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    int a, b;

    if(offs+2!=[args count]){
        errmsg = @"remainder takes exactly two arguments";
        return NO;
    }

    if(nt!=NT_INTEGERS){
        errmsg = @"Arguments to remainder must be integers";
        return NO;
    }

    a = [[args objectAtIndex:offs] intVal];
    b = [[args objectAtIndex:offs+1] intVal];

    if(!b){
        errmsg = @"divide by zero error in remainder";
        return NO;
    }

    value = [[Int alloc] initSCMInt:(a%b)];
    return YES;
}

@end

@implementation PRMList

- (NSString *)primName
{
    return @"list";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id res = [NSNull null];
    int pos;
    
    for(pos=[args count]-1; pos>=offs; pos--){
        res = [Pair newCar:[args objectAtIndex:pos] Cdr:res];
    }

    value = res;
    return YES;
}

@end

@implementation PRMCons

- (NSString *)primName
{
    return @"cons";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    if(offs+2!=[args count]){
        errmsg = @"cons takes exactly two arguments";
        return NO;
    }

    value = [Pair 
                newCar:[args objectAtIndex:offs] 
                Cdr:[args objectAtIndex:offs+1]];
    return YES;
}

@end

@implementation PRMCar

- (NSString *)primName
{
    return @"car";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id item;
    
    if(offs+1!=[args count]){
        errmsg = @"car takes exactly one argument";
        return NO;
    }

    item = [args objectAtIndex:offs];
    if(isPair(item)==NO){
        errmsg = @"argument to car must be a pair";
        return NO;
    }

    value = [item car];
    return YES;
}

@end

@implementation PRMCdr

- (NSString *)primName
{
    return @"cdr";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id item;
    
    if(offs+1!=[args count]){
        errmsg = @"cdr takes exactly one argument";
        return NO;
    }

    item = [args objectAtIndex:offs];
    if(isPair(item)==NO){
        errmsg = @"argument to cdr must be a pair";
        return NO;
    }

    value = [item cdr];
    return YES;
}

@end

@implementation PRMSetCar

- (NSString *)primName
{
    return @"set-car!";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id item;
    
    if(offs+2!=[args count]){
        errmsg = @"set-car! takes two arguments";
        return NO;
    }

    item = [args objectAtIndex:offs];
    if(isPair(item)==NO){
        errmsg = @"argument to set-car! must be a pair";
        return NO;
    }

    [item setcar:[args objectAtIndex:(offs+1)]];

    value = [item car];
    return YES;
}

@end

@implementation PRMSetCdr

- (NSString *)primName
{
    return @"set-cdr!";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id item;
    
    if(offs+2!=[args count]){
        errmsg = @"set-cdr! takes two arguments";
        return NO;
    }

    item = [args objectAtIndex:offs];
    if(isPair(item)==NO){
        errmsg = @"argument to set-cdr! must be a pair";
        return NO;
    }

    [item setcdr:[args objectAtIndex:(offs+1)]];

    value = [item cdr];
    return YES;
}

@end

@implementation PRMDisplay

- (NSString *)primName
{
    return @"display";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id item;
    NSString *res;

    if(offs+1!=[args count]){
        errmsg = @"display takes one argument";
        return NO;
    }

    item = [args objectAtIndex:offs];

    if(isString(item)){
        res = [[NSString alloc]
                  initWithFormat:@"%@" locale: nil,
                  [item strVal]];
    }
    else if(isChar(item)){
        res = [[NSString alloc]
                  initWithFormat:@"%c" locale: nil,
                  [item charVal]];
    }
    else{
        res = [VScheme valToString:item];
    }
    [vm appendToOutput:res];

    value = [NSNull null];
    return YES;
}

@end

@implementation PRMNewline

- (NSString *)primName
{
    return @"newline";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id item;
    
    if(offs!=[args count]){
        errmsg = @"newline takes no arguments";
        return NO;
    }

    [vm appendToOutput:@"\n"];

    value = [NSNull null];
    return YES;
}

@end

@implementation PRMDrawMove
- (NSString *)primName
{
    return @"draw-move";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    DrawInst inst;

    if(offs+2!=[args count]){
        errmsg = @"draw-move takes two arguments";
        return NO;
    }

    if(nt==NT_OTHER){
        errmsg = @"Arguments to draw-move must be numeric";
        return NO;
    }

    inst.what = DRAW_MOVE;
    inst.data.coord.x = [[args objectAtIndex:offs] doubleVal];
    inst.data.coord.y = [[args objectAtIndex:offs+1] doubleVal];

    [vm recordImgInst:inst];

    value = [NSNull null];
    return YES;
}

@end

@implementation PRMDrawLine
- (NSString *)primName
{
    return @"draw-line";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    DrawInst inst;

    if(offs+2!=[args count]){
        errmsg = @"draw-line takes two arguments";
        return NO;
    }

    if(nt==NT_OTHER){
        errmsg = @"Arguments to draw-line must be numeric";
        return NO;
    }

    inst.what = DRAW_LINE;
    inst.data.coord.x = [[args objectAtIndex:offs] doubleVal];
    inst.data.coord.y = [[args objectAtIndex:offs+1] doubleVal];

    [vm recordImgInst:inst];

    value = [NSNull null];
    return YES;
}

@end

@implementation PRMDrawColor
- (NSString *)primName
{
    return @"draw-color";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    DrawInst inst;

    if(offs+3!=[args count]){
        errmsg = @"draw-color takes three arguments";
        return NO;
    }

    if(nt==NT_OTHER){
        errmsg = @"Arguments to draw-color must be numeric";
        return NO;
    }

    inst.what = DRAW_COLOR;
    inst.data.color[0] = [[args objectAtIndex:offs] doubleVal];
    inst.data.color[1] = [[args objectAtIndex:offs+1] doubleVal];
    inst.data.color[2] = [[args objectAtIndex:offs+2] doubleVal];

    if(inst.data.color[0]<(float)0 || 
       inst.data.color[0]>(float)255){
        errmsg = @"red color component out of range (0..255)";
        return NO;
    }
    if(inst.data.color[1]<(float)0 || 
       inst.data.color[1]>(float)255){
        errmsg = @"green color component out of range (0..255)";
        return NO;
    }
    if(inst.data.color[2]<(float)0 || 
       inst.data.color[2]>(float)255){
        errmsg = @"blue color component out of range (0..255)";
        return NO;
    }

    inst.data.color[0]/=(float)255;
    inst.data.color[1]/=(float)255;
    inst.data.color[2]/=(float)255;

    [vm recordImgInst:inst];

    value = [NSNull null];
    return YES;
}

@end

@implementation PRMSin
- (NSString *)primName
{
    return @"sin";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    double res;

    if(offs+1!=[args count]){
        errmsg = @"sin takes one argument";
        return NO;
    }

    if(nt==NT_OTHER){
        errmsg = @"Argument to sin must be numeric";
        return NO;
    }

    res = sin([[args objectAtIndex:offs] doubleVal]);
    value = [[Double alloc] initSCMDouble:res];
    return YES;
}

@end

@implementation PRMASin
- (NSString *)primName
{
    return @"asin";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    double res, v;

    if(offs+1!=[args count]){
        errmsg = @"asin takes one argument";
        return NO;
    }

    if(nt==NT_OTHER){
        errmsg = @"Argument to asin must be numeric";
        return NO;
    }

    v = [[args objectAtIndex:offs] doubleVal];
    if(v<(double)-1.0 || (double)1.0<v){
        errmsg = @"Argument to asin out of range";
        return NO;
    }

    res = asin(v);
    value = [[Double alloc] initSCMDouble:res];
    return YES;
}

@end

@implementation PRMCos
- (NSString *)primName
{
    return @"cos";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    double res;

    if(offs+1!=[args count]){
        errmsg = @"cos takes one argument";
        return NO;
    }

    if(nt==NT_OTHER){
        errmsg = @"Argument to cos must be numeric";
        return NO;
    }

    res = cos([[args objectAtIndex:offs] doubleVal]);
    value = [[Double alloc] initSCMDouble:res];
    return YES;
}

@end

@implementation PRMACos
- (NSString *)primName
{
    return @"acos";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    double res, v;

    if(offs+1!=[args count]){
        errmsg = @"acos takes one argument";
        return NO;
    }

    if(nt==NT_OTHER){
        errmsg = @"Argument to acos must be numeric";
        return NO;
    }

    v = [[args objectAtIndex:offs] doubleVal];
    if(v<(double)-1.0 || (double)1.0<v){
        errmsg = @"Argument to acos out of range";
        return NO;
    }

    res = acos(v);
    value = [[Double alloc] initSCMDouble:res];
    return YES;
}

@end

@implementation PRMSqrt
- (NSString *)primName
{
    return @"sqrt";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    NUMTYPE nt = [self checkArgsNumeric:args offset:offs];
    double v, res; int ires;
    id item;

    if(offs+1!=[args count]){
        errmsg = @"sqrt takes one argument";
        return NO;
    }

    if(nt==NT_OTHER){
        errmsg = @"Argument to sqrt must be numeric";
        return NO;
    }

    item = [args objectAtIndex:offs]; v = [item doubleVal];

    if(v<0){
        errmsg = @"Argument to sqrt must not be negative";
        return NO;
    }

    res  = sqrt([item doubleVal]); 
    ires = (int)res;

    if(nt==NT_INTEGERS && ires*ires==[item intVal]){
        value = [[Int alloc] initSCMInt:ires];
    }
    else{
        value = [[Double alloc] initSCMDouble:res];
    }

    return YES;
}

@end

@implementation PRMMakeVector
- (NSString *)primName
{
    return @"make-vector";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id filler;
    id count;
    int cval;

    if(offs+1==[args count]){
        filler = [NSNull null];
    }
    else if(offs+2==[args count]){
        filler = [args objectAtIndex:offs+1];
    }
    else{
        errmsg = @"make-vector takes one or two arguments";
        return NO;
    }

    count = [args objectAtIndex:offs];

    if(isInt(count)==NO){
        errmsg = @"make-vector: integer required";
        return NO;
    }

    cval = [count intVal];

    if(cval<0){
        errmsg = @"make-vector: non-negative integer required";
        return NO;
    }

    value = [Vector newWithItem:filler count:cval];

    return YES;
}


@end

@implementation PRMListToVector
- (NSString *)primName
{
    return @"list->vector";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id list;

    if(offs+1!=[args count]){
        errmsg = @"list->vector takes one argument";
        return NO;
    }

    list = [args objectAtIndex:offs];

    if(isPair(list)==NO && list!=[NSNull null]){
        errmsg = @"list->vector: list required";
        return NO;
    }

    value = [Vector newFromList:list];

    return YES;
}


@end

@implementation PRMVectorToList
- (NSString *)primName
{
    return @"vector->list";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    Vector *vect;
    Pair *result;
    id *data;
    int index;

    if(offs+1!=[args count]){
        errmsg = @"vector->list takes one argument";
        return NO;
    }

    vect = [args objectAtIndex:offs];

    if(isVector(vect)==NO){
        errmsg = @"vector->list: vector required";
        return NO;
    }

    data = [vect entries];

    for(index=[vect count]-1, result=(Pair *)[NSNull null];
        index>=0; index--){
        result = 
            [Pair newCar:data[index] Cdr:result];
    }

    value = result;

    return YES;
}

@end

@implementation PRMVectorLength
- (NSString *)primName
{
    return @"vector-length";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    Vector *vect;

    if(offs+1!=[args count]){
        errmsg = @"vector-length takes one argument";
        return NO;
    }

    vect = [args objectAtIndex:offs];

    if(isVector(vect)==NO){
        errmsg = @"vector-length: vector required";
        return NO;
    }

    value = [[Int alloc] initSCMInt:[vect count]];

    return YES;
}

@end

@implementation PRMVectorRef
- (NSString *)primName
{
    return @"vector-ref";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id *data;
    Vector *vect;
    int count, index;
    Int *ind;

    if(offs+2!=[args count]){
        errmsg = @"vector-ref takes two arguments";
        return NO;
    }

    vect = [args objectAtIndex:offs];
    if(isVector(vect)==NO){
        errmsg = @"vector-ref: vector required";
        return NO;
    }

    data = [vect entries]; count = [vect count];

    ind = [args objectAtIndex:offs+1];
    if(isInt(ind)==NO){
        errmsg = @"vector-ref: integer required";
        return NO;
    }

    index = [ind intVal];
    if(index<0 || index>=count){
        errmsg = @"vector-ref: index out of bounds";
        return NO;
    }

    value = data[index];

    return YES;
}

@end

@implementation PRMVectorSet
- (NSString *)primName
{
    return @"vector-set!";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id *data;
    Vector *vect;
    int count, index;
    Int *ind;
    id obj;

    if(offs+3!=[args count]){
        errmsg = @"vector-set! takes three arguments";
        return NO;
    }

    vect = [args objectAtIndex:offs];
    if(isVector(vect)==NO){
        errmsg = @"vector-set!: vector required";
        return NO;
    }

    data = [vect entries]; count = [vect count];

    ind = [args objectAtIndex:offs+1];
    if(isInt(ind)==NO){
        errmsg = @"vector-set!: integer required";
        return NO;
    }

    index = [ind intVal];
    if(index<0 || index>=count){
        errmsg = @"vector-set!: index out of bounds";
        return NO;
    }

    obj = [args objectAtIndex:offs+2];

    data[index] = obj; [obj retain];

    value = obj;

    return YES;
}

@end

@implementation PRMVectorFill
- (NSString *)primName
{
    return @"vector-fill!";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id *data;
    Vector *vect;
    int count, index;
    id obj;

    if(offs+2!=[args count]){
        errmsg = @"vector-fill! takes two arguments";
        return NO;
    }

    vect = [args objectAtIndex:offs];
    if(isVector(vect)==NO){
        errmsg = @"vector-fill!: vector required";
        return NO;
    }

    data = [vect entries]; count = [vect count];

    obj = [args objectAtIndex:offs+1];
    for(index=0; index<count; index++){
        data[index] = obj; [obj retain];
    }

    value = obj;

    return YES;
}

@end

@implementation PRMBrowseEnvironment
- (NSString *)primName
{
    return @"browse-environment";
}

- (BOOL)evalVM:(id)vm Args:(NSMutableArray *)args offset:(int)offs
{
    id win;

    if(offs!=[args count]){
        errmsg = @"browse-environment takes no arguments";
        return NO;
    }

    win = [[EnvWindow alloc] initWithEnv:[[vm envStack] lastObject]];
    [[vm delegate] envWindow:win];

    value = [NSNull null];
    return YES;
}

@end
